from __future__ import annotations

import logging
from collections import deque
from typing import Optional

import pyspark.sql.functions as sfn
import wedata.tempo.tsdf as t_tsdf
from pyspark.sql import SparkSession
from pyspark.sql.utils import ParseException

logger = logging.getLogger(__name__)


def write(
    tsdf: t_tsdf.TSDF,
    spark: SparkSession,
    tabName: str,
    optimizationCols: Optional[list[str]] = None,
) -> None:
    """
    param: tsdf: input TSDF object to write
    param: tabName Delta output table name
    param: optimizationCols list of columns to optimize on (time)
    """
    # hilbert curves more evenly distribute performance for querying multiple columns for Delta tables
    spark.conf.set("spark.wedata.io.skipping.mdc.curve", "hilbert")

    df = tsdf.df
    ts_col = tsdf.ts_col
    partitionCols = tsdf.partitionCols

    view_df = df.withColumn("event_dt", sfn.to_date(sfn.col(ts_col))).withColumn(
        "event_time",
        sfn.translate(sfn.split(sfn.col(ts_col).cast("string"), " ")[1], ":", "").cast(
            "double"
        ),
    )
    view_cols = deque(view_df.columns)
    view_cols.rotate(1)
    view_df = view_df.select(*list(view_cols))

    # Use replaceWhere instead of overwrite mode
    writer = view_df.write.format("delta").partitionBy("event_dt")
    writer = writer.option("replaceWhere", "true")
    writer.saveAsTable(tabName)

    if optimizationCols:
        try:
            spark.sql(
                "optimize {} zorder by {}".format(
                    tabName,
                    "(" + ",".join(partitionCols + optimizationCols + [ts_col]) + ")",
                )
            )
        except ParseException as e:
            logger.error(
                "Delta optimizations attempted, but was not successful.\nError: {}".format(
                    e
                )
            )


import copy
import datetime
import logging
import re
from collections import defaultdict
from functools import reduce
from typing import Dict, List, Optional, Tuple

from pyspark.sql import DataFrame
from pyspark.sql import functions as F
import pyspark.sql.functions as psf


from wedata.feature_store.entities.environment_variables import BROADCAST_JOIN_THRESHOLD
from wedata.feature_store.entities.feature_column_info import FeatureColumnInfo
from wedata.feature_store.entities.feature_lookup import FeatureLookup
from wedata.feature_store.entities.feature_spec import FeatureSpec
from wedata.feature_store.entities.feature_table import FeatureTable

from wedata.feature_store.utils import uc_utils

_logger = logging.getLogger(__name__)


def _spark_asof_join_features(
    df: DataFrame,
    df_lookup_keys: List[str],
    df_timestamp_lookup_key: str,
    feature_table_data: DataFrame,
    feature_table_keys: List[str],
    feature_table_timestamp_key: str,
    feature_to_output_name: Dict[str, str],
    lookback_window_seconds: Optional[float] = None,
    use_spark_native_join: Optional[bool] = False,
) -> DataFrame:
    # Alias feature table's keys to DataFrame lookup keys
    ft_key_aliases = [
        feature_table_data[ft_key].alias(df_key)
        for (ft_key, df_key) in zip(feature_table_keys, df_lookup_keys)
    ]
    # Alias features to corresponding output names
    ft_features = [
        (feature_name, output_name)
        for feature_name, output_name in feature_to_output_name.items()
        # Skip join if feature it is already in DataFrame and therefore overridden
        if output_name not in df.columns
    ]
    ft_feature_aliases = [
        feature_table_data[feature_name].alias(output_name)
        for feature_name, output_name in ft_features
    ]
    # Alias feature table's timestamp key to DataFrame timestamp lookup keys
    ft_timestamp_key_aliases = [
        feature_table_data[feature_table_timestamp_key].alias(df_timestamp_lookup_key)
    ]
    # Select key, timestamp key, and feature columns from feature table
    feature_and_keys = feature_table_data.select(
        ft_key_aliases + ft_timestamp_key_aliases + ft_feature_aliases
    )

    _logger.debug(
        "Using native spark for point in time join"
        if use_spark_native_join
        else "Using tempo for point in time join"
    )

    if use_spark_native_join:
        joined_df = _spark_asof_join_features_native(
            labels_df=df,
            features_df=feature_and_keys,
            primary_keys=df_lookup_keys,
            timestamp_key=df_timestamp_lookup_key,
            lookback_window_seconds=lookback_window_seconds,
        )
    else:
        joined_df = _spark_asof_join_features_tempo(
            df=df,
            feature_df=feature_and_keys,
            lookup_keys=df_lookup_keys,
            timestamp_key=df_timestamp_lookup_key,
            lookback_window=lookback_window_seconds,
            ft_features=ft_features,
        )
    return joined_df


def _spark_asof_join_features_tempo(
        df: DataFrame,
        feature_df: DataFrame,
        lookup_keys: List[str],
        timestamp_key: str,
        ft_features: List[Tuple[str, str]],
        lookback_window: Optional[float] = None
) -> DataFrame:
    """
    自定义实现as-of连接
    :param df: 主表DataFrame
    :param feature_df: 特征表DataFrame
    :param lookup_keys: 连接键列表
    :param timestamp_key: 时间戳列名
    :param lookback_window: 最大回溯时间(秒)
    :return: 连接后的DataFrame
    """
    from wedata.tempo.tsdf import TSDF
    # 1. 只保留键列和时间戳列
    df_tsdf = TSDF(df, ts_col=timestamp_key, partition_cols=lookup_keys)
    ft_tsdf = TSDF(feature_df, ts_col=timestamp_key, partition_cols=lookup_keys)
    # 进行as-of连接
    joined_df = df_tsdf.asofJoin(
        ft_tsdf,
        left_prefix="left",
        right_prefix="right",
        skipNulls=False,
        tolerance=lookback_window
        if lookback_window is not None
        else None,
    ).df

    # 去掉前缀，恢复列名
    left_aliases = [
        joined_df[f"left_{column_name}"].alias(column_name)
        for column_name in df.columns
        if column_name not in lookup_keys
    ]
    right_aliases = [
        joined_df[f"right_{output_name}"].alias(output_name)
        for (_, output_name) in ft_features
    ]
    return joined_df.select(lookup_keys + left_aliases + right_aliases)


def _spark_asof_join_features_native(
    labels_df: DataFrame,
    features_df: DataFrame,
    primary_keys: List[str],
    timestamp_key: str,
    lookback_window_seconds: Optional[float] = None,
):
    """
    Performs an as-of join operation between two dataframes using native Spark operations.
    Uses broadcast join for label dataset when within a size threshold to improve
    efficiency of join operation with the assumption that size(labels_df) << size(features_df).
    TODO(ML-40580): automatically switch labels_df and features_df based on size
    The join operation is performed as follows:
        1. Drop non-join key (primary and timestamp keys) columns from labels and features DataFrames
        2. Broadcast join labels onto features DataFrame if within broadcast threshold.
        3. Select maximum timestamp for each primary key
        4. Rejoin non-primary key columns from features DataFrame to get features data
        5. Rejoin non-primary key columns from labels DataFrame to get joint data

    Parameters:
    labels_df (DataFrame): The labels dataframe to join.
    features_df (DataFrame): The features dataframe to join.
    primary_keys (List[str]): The primary keys used for joining.
    timestamp_key (str): The timestamp key used for joining.
    lookback_window_seconds (Optional[float]): The lookback window in seconds.
        If provided, the join operation will only consider records within this window.

    Returns:
    DataFrame: The result of the as-of join operation.
    """
    labels_df_keys_only = labels_df.select(
        [F.col(key) for key in primary_keys] + [F.col(timestamp_key)]
    )

    # Broadcast labels DataFrame if within the broadcast threshold
    if _df_in_size_threshold(labels_df_keys_only, BROADCAST_JOIN_THRESHOLD.get()):
        labels_df_keys_only = F.broadcast(labels_df_keys_only)

    # Drop non-primary key columns from features DataFrame
    features_df_keys_only = features_df.select(
        [F.col(key).alias(f"__features_pk_{key}") for key in primary_keys]
        + [F.col(timestamp_key).alias("__features_tk")]
    )

    # Create join conditions
    join_conditions = [
        labels_df_keys_only[key] == features_df_keys_only[f"__features_pk_{key}"]
        for key in primary_keys
    ]
    join_conditions = reduce(lambda x, y: x & y, join_conditions)
    join_conditions &= (
        labels_df_keys_only[timestamp_key] >= features_df_keys_only["__features_tk"]
    )
    if lookback_window_seconds is not None:
        join_conditions &= (
            psf.unix_timestamp(labels_df_keys_only[timestamp_key])
            - psf.unix_timestamp(features_df_keys_only["__features_tk"])
        ) <= lookback_window_seconds

    # Join labels and features DataFrames
    labels_df_keys_with_features_keys = labels_df_keys_only.join(
        features_df_keys_only, on=join_conditions, how="left"
    )

    # Find the features max timestamps for each primary keys and timestamp key in labels
    labels_df_keys_with_features_keys = labels_df_keys_with_features_keys.groupBy(
        [labels_df_keys_only[key] for key in primary_keys] + [F.col(timestamp_key)]
    ).agg(F.max("__features_tk").alias("__max_ts"))

    if _df_in_size_threshold(
        labels_df_keys_with_features_keys, BROADCAST_JOIN_THRESHOLD.get()
    ):
        labels_df_keys_with_features_keys = F.broadcast(
            labels_df_keys_with_features_keys
        )

    # Rejoin features DataFrame to get the features data
    join_conditions = [
        features_df[key] == labels_df_keys_with_features_keys[key]
        for key in primary_keys
    ]
    join_conditions = reduce(lambda x, y: x & y, join_conditions)
    join_conditions &= (
        features_df[timestamp_key] == labels_df_keys_with_features_keys["__max_ts"]
    )

    features = features_df.join(
        labels_df_keys_with_features_keys,
        on=join_conditions,
        how="inner",
    )

    pk_columns_to_drop = [
        labels_df_keys_with_features_keys[key] for key in primary_keys
    ]
    features = features.drop(*pk_columns_to_drop).drop(
        features_df[timestamp_key], labels_df_keys_with_features_keys["__max_ts"]
    )
    features = features.dropDuplicates(primary_keys + [timestamp_key])
    # Rejoin labels DataFrame if columns were dropped
    joint_df = labels_df.join(features, on=primary_keys + [timestamp_key], how="left")
    return joint_df


def _df_in_size_threshold(df, threshold) -> float:
    # Default to within threshold if can not find
    try:
        num_bytes = _get_df_size_from_spark_plan(df)
    except Exception as e:
        num_bytes = 0
    return num_bytes <= threshold


def _get_df_size_from_spark_plan(df: DataFrame) -> float:
    """
    获取DataFrame的估算大小(字节数)
    替代方案：直接从DataFrame的SparkSession获取执行计划信息

    参数:
        df: 要计算大小的Spark DataFrame

    返回:
        float: 估算的字节数

    异常:
        ValueError: 如果无法从执行计划中解析出大小信息
    """
    # 直接从DataFrame获取SparkSession
    spark = df.sql_ctx.sparkSession

    # 创建临时视图
    df.createOrReplaceTempView("temp_view_for_size")

    # 获取执行计划
    plan = spark.sql("explain cost select * from temp_view_for_size").collect()[0][0]

    # 解析大小信息
    search_result = re.search(r"sizeInBytes=.*(['\)])", plan, re.MULTILINE)
    if search_result is None:
        raise ValueError("无法从Spark执行计划中获取sizeInBytes信息")

    # 提取大小和单位
    result = search_result.group(0).replace(")", "")
    size, units = result.split("=")[1].split()

    # 单位转换映射
    units_map = {
        "TiB": 1024**4,  # 太字节
        "GiB": 1024**3,  # 吉字节
        "MiB": 1024**2,  # 兆字节
        "KiB": 1024,     # 千字节
        "B": 1           # 字节(处理没有单位的情况)
    }

    # 清理单位字符串并转换
    clean_units = units.rstrip(",")
    return float(size) * units_map.get(clean_units, 1)  # 默认返回原始值


def _spark_join_features(
    df: DataFrame,
    df_keys: List[str],
    feature_table_data: DataFrame,
    feature_table_keys: List[str],
    feature_to_output_name: Dict[str, str],
) -> DataFrame:
    """
    Helper to join `feature_name` from `feature_table_data` into `df`.

    This join uses a temporary table that contains only the keys and feature
    from the feature table. The temporary table aliases the keys to match
    the lookup keys and the feature to match the output_name.

    Aliasing the keys allows us to join on name instead of by column,
    which prevents duplicate column names after the join.
    (see: https://kb.databricks.com/data/join-two-dataframes-duplicated-columns.html)

    The joined-in feature is guaranteed to be unique because FeatureSpec
    columns must be unique and the join is skipped if the feature
    already exists in the DataFrame.
    """

    # Alias feature table's keys to DataFrame lookup keys
    ft_key_aliases = [
        feature_table_data[ft_key].alias(df_key)
        for (ft_key, df_key) in zip(feature_table_keys, df_keys)
    ]
    # Alias features to corresponding output names
    ft_feature_aliases = [
        feature_table_data[feature_name].alias(output_name)
        for feature_name, output_name in feature_to_output_name.items()
        # Skip join if feature it is already in DataFrame and therefore overridden
        if output_name not in df.columns
    ]
    # Select key and feature columns from feature table
    feature_and_keys = feature_table_data.select(ft_key_aliases + ft_feature_aliases)
    # Join feature to feature table
    return df.join(feature_and_keys, df_keys, how="left")


def _validate_join_keys(
    feature_column_info: FeatureColumnInfo,
    df: DataFrame,
    feature_table_metadata: FeatureTable,
    feature_table_data: DataFrame,
    is_timestamp_key: bool = False,
):
    join_error_phrase = (
        f"Unable to join feature table '{feature_column_info.table_name}'"
    )
    feature_column_info_keys = (
        feature_column_info.timestamp_lookup_key
        if is_timestamp_key
        else feature_column_info.lookup_key
    )
    feature_table_keys = (
        feature_table_metadata.timestamp_keys
        if is_timestamp_key
        else feature_table_metadata.primary_keys
    )

    lookup_key_kind = "timestamp lookup key" if is_timestamp_key else "lookup key"
    feature_table_key_kind = "timestamp key" if is_timestamp_key else "primary key"

    # Validate df has necessary keys
    missing_df_keys = list(
        filter(lambda df_key: df_key not in df.columns, feature_column_info_keys)
    )
    if missing_df_keys:
        missing_keys = ", ".join([f"'{key}'" for key in missing_df_keys])
        raise ValueError(
            f"{join_error_phrase} because {lookup_key_kind} {missing_keys} not found in DataFrame."
        )
    # Validate feature table has necessary keys
    missing_ft_keys = list(
        filter(
            lambda ft_key: ft_key not in feature_table_data.columns, feature_table_keys
        )
    )
    if missing_ft_keys:
        missing_keys = ", ".join([f"'{key}'" for key in missing_ft_keys])
        raise ValueError(
            f"{join_error_phrase} because {feature_table_key_kind} {missing_keys} not found in feature table."
        )

    # Validate number of feature table keys matches number of df lookup keys
    if len(feature_column_info_keys) != len(feature_table_keys):
        raise ValueError(
            f"{join_error_phrase} because "
            f"number of {feature_table_key_kind}s ({feature_table_keys}) "
            f"does not match "
            f"number of {lookup_key_kind}s ({feature_column_info_keys})."
        )

    # Validate feature table keys match types of df keys. The number of keys is expected to be the same.
    # for (df_key, ft_key) in zip(feature_column_info_keys, feature_table_keys):
    #     df_key_type = DataType.from_spark_type(df.schema[df_key].dataType)
    #     ft_key_type = DataType.from_spark_type(
    #         feature_table_data.schema[ft_key].dataType
    #     )
    #     if df_key_type != ft_key_type:
    #         raise ValueError(
    #             f"{join_error_phrase} because {feature_table_key_kind} '{ft_key}' has type '{DataType.to_string(ft_key_type)}' "
    #             f"but corresponding {lookup_key_kind} '{df_key}' has type '{DataType.to_string(df_key_type)}' in DataFrame."
    #         )


def _validate_join_feature_data(
    df: DataFrame,
    features_to_join: List[FeatureColumnInfo],
    feature_table_metadata_map: Dict[str, FeatureTable],
    feature_table_data_map: Dict[str, DataFrame],
):
    for feature_info in features_to_join:
        feature_table_metadata = feature_table_metadata_map[feature_info.table_name]
        feature_table_data = feature_table_data_map[feature_info.table_name]
        # Validate feature table primary keys match length/type of df lookup keys
        _validate_join_keys(
            feature_info,
            df,
            feature_table_metadata,
            feature_table_data,
            is_timestamp_key=False,
        )
        # Validate feature table timestamp keys match length/type of df timestamp lookup keys
        _validate_join_keys(
            feature_info,
            df,
            feature_table_metadata,
            feature_table_data,
            is_timestamp_key=True,
        )


def join_feature_data_if_not_overridden(
    feature_spec: FeatureSpec,
    df: DataFrame,
    features_to_join: List[FeatureColumnInfo],
    feature_table_metadata_map: Dict[str, FeatureTable],
    feature_table_data_map: Dict[str, DataFrame],
    use_spark_native_join: Optional[bool] = False,
) -> DataFrame:
    """
    Joins `df` with features specified by `feature_spec.feature_column_infos` if they do not already exist.

    Return column order is df.columns + newly joined features. The newly joined feature order is not guaranteed to
    match `feature_spec.feature_column_infos` as feature lookups are first grouped by table for efficiency.

    Before joining, it checks that:
    1. Feature table keys match length and types of `df` lookup keys specified by FeatureSpec
    2. `df` contains lookup keys specified by FeatureSpec
    3. Feature table timestamp lookup keys match length and types of `df` timestamp lookup keys if specified by FeatureSpec
    4. `df` contains timestamp lookup keys if specified by FeatureSpec
    """
    _validate_join_feature_data(
        df=df,
        features_to_join=features_to_join,
        feature_table_metadata_map=feature_table_metadata_map,
        feature_table_data_map=feature_table_data_map,
    )

    # Helper class to group all unique combinations of feature table names and lookup keys.
    # All features in each of these groups will be JOINed with the training df using a single JOIN.
    class JoinDataKey:
        def __init__(
            self,
            feature_table: str,
            lookup_key: List[str],
            timestamp_lookup_key: List[str],
            lookback_window: Optional[datetime.timedelta] = None,
        ):
            self.feature_table = feature_table
            self.lookup_key = lookup_key
            self.timestamp_lookup_key = timestamp_lookup_key
            self.lookback_window = lookback_window

        def __hash__(self):
            return (
                hash(self.feature_table)
                + hash(tuple(self.lookup_key))
                + hash(tuple(self.timestamp_lookup_key))
                + hash(self.lookback_window)
            )

        def __eq__(self, other):
            return (
                self.feature_table == other.feature_table
                and self.lookup_key == other.lookup_key
                and self.timestamp_lookup_key == other.timestamp_lookup_key
                and self.lookback_window == other.lookback_window
            )

    # Iterate through the list of FeatureColumnInfo and group features by name of the
    # feature table and lookup key(s) and timestamp lookup key(s)
    table_join_data = defaultdict(dict)
    lookback_windows = {
        t.table_name: t.lookback_window for t in feature_spec.table_infos
    }
    for feature_info in features_to_join:
        join_data_key = JoinDataKey(
            feature_info.table_name,
            feature_info.lookup_key,
            feature_info.timestamp_lookup_key,
            lookback_windows[feature_info.table_name],
        )
        table_join_data[join_data_key][
            feature_info.feature_name
        ] = feature_info.output_name

    for join_data_key, feature_to_output_name in table_join_data.items():

        feature_table_metadata = feature_table_metadata_map[join_data_key.feature_table]
        feature_table_data = feature_table_data_map[join_data_key.feature_table]

        if join_data_key.timestamp_lookup_key:
            # If lookback window is set to 0, then perform exact join instead of asof join to get perf benefits.
            if (
                join_data_key.lookback_window is not None
                and join_data_key.lookback_window == 0
            ):
                df = _spark_join_features(
                    df=df,
                    df_keys=join_data_key.lookup_key
                    + join_data_key.timestamp_lookup_key,
                    feature_table_data=feature_table_data,
                    feature_table_keys=feature_table_metadata.primary_keys
                    + feature_table_metadata.timestamp_keys,
                    feature_to_output_name=feature_to_output_name,
                )
            else:
                df = _spark_asof_join_features(
                    df=df,
                    df_lookup_keys=join_data_key.lookup_key,
                    df_timestamp_lookup_key=join_data_key.timestamp_lookup_key[0],
                    feature_table_data=feature_table_data,
                    feature_table_keys=feature_table_metadata.primary_keys,
                    feature_table_timestamp_key=feature_table_metadata.timestamp_keys[
                        0
                    ],
                    feature_to_output_name=feature_to_output_name,
                    lookback_window_seconds=join_data_key.lookback_window,
                    use_spark_native_join=use_spark_native_join,
                )
        else:
            df = _spark_join_features(
                df=df,
                df_keys=join_data_key.lookup_key,
                feature_table_data=feature_table_data,
                feature_table_keys=feature_table_metadata.primary_keys,
                feature_to_output_name=feature_to_output_name,
            )
    return df


def get_feature_lookups_with_full_table_names(
        feature_lookups: List[FeatureLookup], current_catalog: str, current_schema: str
) -> List[FeatureLookup]:
    """
    Takes in a list of FeatureLookups, and returns copies with reformatted table names.
    """
    table_names = {fl.table_name for fl in feature_lookups}
    uc_utils._check_qualified_table_names(table_names)
    uc_utils._verify_all_tables_are_either_in_uc_or_in_hms(
        table_names, current_catalog, current_schema
    )
    standardized_feature_lookups = []
    for fl in feature_lookups:
        fl_copy = copy.deepcopy(fl)
        fl_copy._table_name = uc_utils.get_full_table_name(
            fl_copy.table_name, current_catalog, current_schema
        )
        standardized_feature_lookups.append(fl_copy)
    return standardized_feature_lookups


import json

from tencentcloud.wedata.v20210820.wedata_client import WedataClient
from tencentcloud.common import credential
from tencentcloud.common.exception.tencent_cloud_sdk_exception import TencentCloudSDKException
from wedata.feature_store.cloud_sdk_client.utils import get_client_profile, set_request_header
import  wedata.feature_store.cloud_sdk_client.models  as models


class FeatureCloudSDK:
    def __init__(self, secret_id: str, secret_key: str, region: str):
        self._client = WedataClient(credential.Credential(secret_id, secret_key), region, get_client_profile())

    def CreateOnlineFeatureTable(self, request: models.CreateOnlineFeatureTableRequest) -> 'models.CreateOnlineFeatureTableResponse':
        """
        创建在线特征表
        Args:
            request: 创建请求参数

        Returns:
            创建结果响应
        """
        try:
            params = request._serialize()
            headers = set_request_header(request.headers)
            print(f"CreateOnlineFeatureTable params: {params}")
            print(f"CreateOnlineFeatureTable headers: {headers}")
            body = self._client.call("CreateOnlineFeatureTable", params, headers=headers)
            response = json.loads(body)
            model = models.CreateOnlineFeatureTableResponse()
            model._deserialize(response["Response"])
            print("CreateOnlineFeatureTable Response: ", response)
            return model
        except Exception as e:
            if isinstance(e, TencentCloudSDKException):
                raise
            else:
                raise TencentCloudSDKException(type(e).__name__, str(e))

    def DescribeNormalSchedulerExecutorGroups(self, request: models.DescribeNormalSchedulerExecutorGroupsRequest) -> 'models.DescribeNormalSchedulerExecutorGroupsResponse':
        """
        查询普通调度器执行器组
        Args:
            request: 查询请求参数

        Returns:
            查询结果响应
        """
        try:
            params = request._serialize()
            headers = set_request_header(request.headers)
            print(f"DescribeNormalSchedulerExecutorGroups params: {params}")
            print(f"DescribeNormalSchedulerExecutorGroups headers: {headers}")
            body = self._client.call("DescribeNormalSchedulerExecutorGroups", params, headers=headers)
            response = json.loads(body)
            model = models.DescribeNormalSchedulerExecutorGroupsResponse()
            model._deserialize(response["Response"])
            print("DescribeNormalSchedulerExecutorGroups Response: ", response)
            return model
        except Exception as e:
            if isinstance(e, TencentCloudSDKException):
                raise
            else:
                raise TencentCloudSDKException(type(e).__name__, str(e))

    def RefreshFeatureTable(self, request: models.RefreshFeatureTableRequest) -> 'models.RefreshFeatureTableResponse':
        """
        刷新特征表
        Args:
            request: 刷新请求参数
        Returns:
            刷新结果响应
        """
        try:
            params = request._serialize()
            headers = set_request_header(request.headers)
            print(f"RefreshFeatureTable params: {params}")
            print(f"RefreshFeatureTable headers: {headers}")
            body = self._client.call("RefreshFeatureTable", params, headers=headers)
            response = json.loads(body)
            model = models.RefreshFeatureTableResponse()
            model._deserialize(response["Response"])
            print("RefreshFeatureTable Response: ", response)
            return model
        except Exception as e:
            if isinstance(e, TencentCloudSDKException):
                raise
            else:
                raise TencentCloudSDKException(type(e).__name__, str(e))


from collections import defaultdict
from typing import Optional, Any, Dict, List

import logging
from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.catalog import Column
from pyspark.sql.functions import when, isnull
from pyspark.sql.types import StructType, StringType, StructField
from mlflow.pyfunc import spark_udf

from wedata.feature_store.constants.constants import (
    _PREBUILT_ENV_URI
)

from wedata.feature_store.entities.feature import Feature
from wedata.feature_store.entities.feature_table import FeatureTable
from wedata.feature_store.entities.function_info import FunctionParameterInfo, FunctionInfo
from wedata.feature_store.utils.common_utils import unsanitize_identifier, check_spark_table_exists, check_package_version


class SparkClient:
    def __init__(self, spark: SparkSession):
        self._spark = spark

    def get_current_catalog(self):
        """
        获取当前Spark会话的catalog名称（使用spark.catalog.currentCatalog属性）

        返回:
            str: 当前catalog名称，如果未设置则返回None
        """
        _, ok, _ = check_package_version("pyspark", "3.4.0", ">=")
        print(f"pyspark >= 3.4.0  {ok}")
        if ok:
            return unsanitize_identifier(self._spark.catalog.currentCatalog())
        else:
            catalog = self._spark.sql("SELECT current_catalog()").first()[0]
            return unsanitize_identifier(catalog)


    def get_current_database(self):
        """
        获取Spark上下文中当前设置的数据库名称

        返回:
            str: 当前数据库名称，如果获取失败则返回None
        """
        try:
            # 使用Spark SQL查询当前数据库
            df = self._spark.sql("SELECT CURRENT_DATABASE()")
            # 获取第一行第一列的值并去除特殊字符
            return unsanitize_identifier(df.first()[0])
        except Exception:
            # 捕获所有异常并返回None
            return None

    def createDataFrame(self, data, schema) -> DataFrame:
        return self._spark.createDataFrame(data, schema)


    def read_table(self, table_name):
        """读取Spark表数据

        Args:
            table_name: 表名，支持格式: catalog.schema.table、schema.table

        Returns:
            DataFrame: 表数据

        Raises:
            ValueError: 当表不存在或读取失败时抛出
        """
        table_name = _translate_spark_table_name(table_name)
        try:
            # 验证表是否存在
            if not  check_spark_table_exists(self._spark, table_name):
                raise ValueError(f"Table does not exist: {table_name}")
            return self._spark.table(table_name)

        except Exception as e:
            raise ValueError(f"Failed to read table {table_name}: {str(e)}")

    def get_features(self, table_name):
        # 查询列信息
        table_name = _translate_spark_table_name(table_name)
        split = table_name.split(".")
        if len(split) == 2:
            # db.table_name
            columns = self._spark.catalog.listColumns(tableName=split[1], dbName=split[0])
        elif len(split) == 3:
            # catalog.db.table_name
            columns = self._spark.catalog.listColumns(tableName=split[2], dbName=split[1])
        else:
            columns = self._spark.catalog.listColumns(tableName=table_name)
        return [
            Feature(
                feature_table=table_name,
                feature_id=f"{table_name}_{row.name}",
                name=row.name,
                data_type=row.dataType,
                description=row.description or ""
            ) for row in columns
        ]

    def get_feature_table(self, table_name):
        """
        DLC支持table_name为catalog.schema.table
        EMR支持table_name为schema.table
        """

        table_name = _translate_spark_table_name(table_name)
        # table = self._spark.catalog.getTable(table_name)
        # 获取表配置信息
        properties = self._spark.sql(f"SHOW TBLPROPERTIES {table_name}").collect()
        primary_key_str = next((row.value for row in properties if row.key == "primaryKeys"), None)
        primary_keys = primary_key_str.split(",") if primary_key_str else []
        table_id = next((row.value for row in properties if row.key == "table_id"), table_name)
        description = next((row.value for row in properties if row.key == "comment"), None)
        timestamp_keys_str = next((row.value for row in properties if row.key == "timestampKeys"), None)
        timestamp_keys = timestamp_keys_str.split(",") if timestamp_keys_str else []
        # 获取分区字段信息
        desc_df = self._spark.sql(f"DESCRIBE EXTENDED {table_name}")
        partition_info = desc_df.filter("col_name LIKE '_partition%'").collect()
        partition_columns = []
        if partition_info:
            partition_str = partition_info[0]["data_type"]
            # 从分区字符串中提取分区字段
            if partition_str.startswith("struct<") and partition_str.endswith(">"):
                # 去掉struct<>外壳
                fields_str = partition_str[7:-1]
                # 分割各个字段定义
                field_defs = [f.strip() for f in fields_str.split(",") if f.strip()]
                # 提取字段名
                partition_columns = [f.split(":")[0].strip() for f in field_defs]

        # 获取特征列信息
        features = self.get_features(table_name)

        # 构建完整的FeatureTable对象
        return FeatureTable(
            name=table_name,
            table_id=table_id,
            description=description,
            primary_keys=primary_keys,
            partition_columns=partition_columns,
            features=features,
            creation_timestamp=None,  # Spark表元数据不包含创建时间戳
            online_stores=None,
            notebook_producers=None,
            job_producers=None,
            table_data_sources=None,
            path_data_sources=None,
            custom_data_sources=None,
            timestamp_keys=timestamp_keys,
            tags=None
        )

    def _get_routines_with_parameters(self, full_routine_names: List[str]) -> DataFrame:
        """
        Retrieve the routines with their parameters from information_schema.routines, information_schema.parameters.
        Return DataFrame only contains routines that 1. exist and 2. the caller has GetFunction permission on.

        Note: The returned DataFrame contains the cartesian product of routines and parameters.
        For efficiency, routines table columns are only present in the first row for each routine.
        """
        routine_name_schema = StructType(
            [
                StructField("specific_catalog", StringType(), False),
                StructField("specific_schema", StringType(), False),
                StructField("specific_name", StringType(), False),
            ]
        )
        routine_names_df = self.createDataFrame(
            [full_routine_name.split(".") for full_routine_name in full_routine_names],
            routine_name_schema,
        )
        routines_table = self.read_table(
            "system.information_schema.routines"
        )
        parameters_table = self.read_table(
            "system.information_schema.parameters"
        )

        # Inner join routines table to filter out non-existent routines.
        # Left join parameters as routines may have no parameters.
        full_routines_with_parameters_df = routine_names_df.join(
            routines_table, on=routine_names_df.columns, how="inner"
        ).join(parameters_table, on=routine_names_df.columns, how="left")

        # Return only relevant metadata from information_schema, sorted by routine name + parameter order.
        # For efficiency, only preserve routine column values in the first of each routine's result rows.
        # The first row will have parameter.ordinal_value is None (no parameters) or equals 0 (first parameter).
        def select_if_first_row(col: Column) -> Column:
            return when(
                isnull(parameters_table.ordinal_position)
                | (parameters_table.ordinal_position == 0),
                col,
                ).otherwise(None)

        return full_routines_with_parameters_df.select(
            routine_names_df.columns
            + [
                select_if_first_row(routines_table.routine_definition).alias(
                    "routine_definition"
                ),
                select_if_first_row(routines_table.external_language).alias(
                    "external_language"
                ),
                parameters_table.ordinal_position,
                parameters_table.parameter_name,
                parameters_table.full_data_type,
            ]
        ).sort(routine_names_df.columns + [parameters_table.ordinal_position])

    def get_functions(self, full_function_names: List[str]) -> List[FunctionInfo]:
        """
        Retrieves and maps Unity Catalog functions' metadata as FunctionInfos.
        """
        # Avoid unnecessary Spark calls and return if empty.
        if not full_function_names:
            return []

        # Collect dict of routine name -> DataFrame rows describing the routine.
        routines_with_parameters_df = self._get_routines_with_parameters(
            full_routine_names=full_function_names
        )
        routine_infos = defaultdict(list)
        for r in routines_with_parameters_df.collect():
            routine_name = f"{r.specific_catalog}.{r.specific_schema}.{r.specific_name}"
            routine_infos[routine_name].append(r)

        # Mock GetFunction DNE error, since information_schema does not throw.
        for function_name in full_function_names:
            if not function_name in routine_infos:
                raise ValueError(f"Function '{function_name}' does not exist.")

        # Map routine_infos into FunctionInfos.
        function_infos = []
        for function_name in full_function_names:
            routine_info = routine_infos[function_name][0]
            input_params = [
                FunctionParameterInfo(name=p.parameter_name, type_text=p.full_data_type)
                for p in routine_infos[function_name]
                if p.ordinal_position is not None
            ]
            function_infos.append(
                FunctionInfo(
                    full_name=function_name,
                    input_params=input_params,
                    routine_definition=routine_info.routine_definition,
                    external_language=routine_info.external_language,
                )
            )
        return function_infos

    def get_predict_udf(
        self,
        model_uri,
        result_type=None,
        env_manager=None,
        params: Optional[dict[str, Any]] = None,
        prebuilt_env_uri: Optional[str] = None,
    ):
        kwargs = {}
        if result_type:
            kwargs["result_type"] = result_type
        if env_manager:
            kwargs["env_manager"] = env_manager
        if params:
            kwargs["params"] = params
        if prebuilt_env_uri:
            kwargs[_PREBUILT_ENV_URI] = prebuilt_env_uri

        return spark_udf(self._spark, model_uri, **kwargs)


def _translate_spark_table_name(table_name):
    from wedata.feature_store.constants.engine_types import judge_engine_type, CalculateEngineTypes
    # 获取表元数据
    if judge_engine_type() == CalculateEngineTypes.EMR:
        split_names = table_name.split(".")
        # print(f"==== EMR TABLE split len({len(split_names)})")
        if len(split_names) <= 2:
            return table_name
        else:
            table_name = ".".join(table_name.split(".")[1:])
            return table_name
    return table_name





OVERWRITE = "overwrite"
APPEND = "append"
PATH = "path"
TABLE = "table"
CUSTOM = "custom"
PREDICTION_COLUMN_NAME = "prediction"
MODEL_DATA_PATH_ROOT = "feature_store"
RAW_MODEL_FOLDER = "raw_model"
UTF8_BYTES_PER_CHAR = 4
MAX_PRIMARY_KEY_STRING_LENGTH_CHARS = 100
MAX_PRIMARY_KEY_STRING_LENGTH_BYTES = (
    MAX_PRIMARY_KEY_STRING_LENGTH_CHARS * UTF8_BYTES_PER_CHAR
)
STREAMING_TRIGGER_CONTINUOUS = "continuous"
STREAMING_TRIGGER_ONCE = "once"
STREAMING_TRIGGER_PROCESSING_TIME = "processingTime"
DEFAULT_WRITE_STREAM_TRIGGER = {STREAMING_TRIGGER_PROCESSING_TIME: "5 seconds"}
_DEFAULT_PUBLISH_STREAM_TRIGGER = {STREAMING_TRIGGER_PROCESSING_TIME: "5 minutes"}
FEATURE_STORE_CLIENT = "FeatureStoreClient"


_WARN = "WARN"
_ERROR = "ERROR"
_SOURCE_FORMAT_DELTA = "delta"

_NO_RESULT_TYPE_PASSED = "NO_RESULT_TYPE"
_USE_SPARK_NATIVE_JOIN = "use_spark_native_join"
_PREBUILT_ENV_URI = "prebuilt_env_uri"

# MLflow模型相关常量(原mlflow_model_constants.py)
# Module name of the original mlflow_model
MLFLOW_MODEL_NAME = "wedata.feature_store.mlflow_model"

# FeatureStoreClient.log_model将记录包含'raw_model'文件夹的模型
# 该文件夹存储原始模型的MLmodel文件，用于推理
RAW_MODEL_FOLDER = "raw_model"

# ML模型文件名常量
ML_MODEL = "MLmodel"

# 特征查找客户端的PyPI包名
FEATURE_LOOKUP_CLIENT_PIP_PACKAGE = "tencent-wedata-feature-engineering-dev"

# 特征查找版本号
FEATURE_LOOKUP_CLIENT_MAJOR_VERSION = "0.1.36"

# 特征存储内部数据目录
FEATURE_STORE_INTERNAL_DATA_DIR = "_wedata_internal/"
WEDATA_DEFAULT_FEATURE_STORE_DATABASE = "WEDATA_DEFAULT_FEATURE_STORE_DATABASE"

# 特征表属性
FEATURE_TABLE_KEY = "wedata.feature_table"
FEATURE_TABLE_VALUE = "true"

FEATURE_TABLE_PROJECT = "wedata.feature_project_id"
FEATURE_TABLE_TIMESTAMP = "timestampKeys"
FEATURE_TABLE_BACKUP_PRIMARY_KEY = "primaryKeys"    # 备用标识，主键
FEATURE_DLC_TABLE_PRIMARY_KEY = "dlc.ao.data.govern.sorted.keys"




from typing import List, Optional

from wedata.feature_store.common.protos import feature_store_pb2

class FunctionParameterInfo():
    def __init__(self, name: str, type_text: str):
        self._name = name
        self._type_text = type_text

    @property
    def name(self) -> str:
        return self._name

    @property
    def type_text(self) -> str:
        return self._type_text

    @classmethod
    def from_dict(cls, function_parameter_info_json):
        return FunctionParameterInfo(
            function_parameter_info_json["name"],
            function_parameter_info_json["type_text"],
        )


class FunctionInfo():
    """
    Helper entity class that exposes properties in GetFunction's response JSON as attributes.
    https://docs.databricks.com/api-explorer/workspace/functions/get

    Note: empty fields (e.g. when 0 input parameters) are not included in the response JSON.
    """

    # Python UDFs have external_language = "Python"
    PYTHON = "Python"

    def __init__(
        self,
        full_name: str,
        input_params: List[FunctionParameterInfo],
        routine_definition: Optional[str],
        external_language: Optional[str],
    ):
        self._full_name = full_name
        self._input_params = input_params
        self._routine_definition = routine_definition
        self._external_language = external_language


    @property
    def full_name(self) -> str:
        return self._full_name

    @property
    def input_params(self) -> List[FunctionParameterInfo]:
        return self._input_params

    @property
    def routine_definition(self) -> Optional[str]:
        return self._routine_definition

    @property
    def external_language(self) -> Optional[str]:
        """
        Field is None if language is SQL (not an external language).
        """
        return self._external_language

    @classmethod
    def from_dict(cls, function_info_json):
        input_params = function_info_json.get("input_params", {}).get("parameters", [])
        return FunctionInfo(
            full_name=function_info_json["full_name"],
            input_params=[FunctionParameterInfo.from_dict(p) for p in input_params],
            routine_definition=function_info_json.get("routine_definition", None),
            external_language=function_info_json.get("external_language", None),
        )

    @classmethod
    def from_proto(cls, function_info_proto):
        return cls(full_name=function_info_proto.udf_name)

    def to_proto(self):
        return feature_store_pb2.FunctionInfo(
            full_name=self.full_name,
            input_params=self.input_params,
            routine_definition=self.routine_definition,
            external_language=self.external_language
        )



import copy
from typing import List, Union

from wedata.feature_store.entities.feature_function import FeatureFunction
from wedata.feature_store.entities.feature_lookup import FeatureLookup
from wedata.feature_store.spark_client.spark_client import SparkClient
from wedata.feature_store.utils import uc_utils
from wedata.feature_store.utils.feature_lookup_utils import get_feature_lookups_with_full_table_names


def format_feature_lookups_and_functions(
        _spark_client: SparkClient, features: List[Union[FeatureLookup, FeatureFunction]]
):
    fl_idx = []
    ff_idx = []
    feature_lookups = []
    feature_functions = []
    for idx, feature in enumerate(features):
        if isinstance(feature, FeatureLookup):
            fl_idx.append(idx)
            feature_lookups.append(feature)
        elif isinstance(feature, FeatureFunction):
            ff_idx.append(idx)
            feature_functions.append(feature)
        else:
            raise ValueError(
                f"Expected a list of FeatureLookups for 'feature_lookups', but received type '{type(feature)}'."
            )

    # FeatureLookups and FeatureFunctions must have fully qualified table, UDF names
    feature_lookups = get_feature_lookups_with_full_table_names(
        feature_lookups,
        _spark_client.get_current_catalog(),
        _spark_client.get_current_database(),
    )
    feature_functions = get_feature_functions_with_full_udf_names(
        feature_functions,
        _spark_client.get_current_catalog(),
        _spark_client.get_current_database(),
    )

    # Restore original order of FeatureLookups, FeatureFunctions. Copy to avoid mutating original list.
    features = features.copy()
    for idx, feature in zip(fl_idx + ff_idx, feature_lookups + feature_functions):
        features[idx] = feature

    return features


def get_feature_functions_with_full_udf_names(
        feature_functions: List[FeatureFunction], current_catalog: str, current_schema: str
):
    """
    Takes in a list of FeatureFunctions, and returns copies with:
    1. Fully qualified UDF names.
    2. If output_name is empty, fully qualified UDF names as output_name.
    """
    udf_names = {ff.udf_name for ff in feature_functions}
    uc_utils._check_qualified_udf_names(udf_names)
    uc_utils._verify_all_udfs_in_uc(udf_names, current_catalog, current_schema)

    standardized_feature_functions = []
    for ff in feature_functions:
        ff_copy = copy.deepcopy(ff)
        del ff

        ff_copy._udf_name = uc_utils.get_full_udf_name(
            ff_copy.udf_name, current_catalog, current_schema
        )
        if not ff_copy.output_name:
            ff_copy._output_name = ff_copy.udf_name
        standardized_feature_functions.append(ff_copy)
    return standardized_feature_functions



from __future__ import annotations

from typing import Callable, List, Optional, Union

from pyspark.sql.dataframe import DataFrame
import pyspark.sql.functions as sfn
from pyspark.sql.types import NumericType
from pyspark.sql.window import Window

import wedata.tempo.resample as t_resample
import wedata.tempo.tsdf as t_tsdf
import wedata.tempo.utils as t_utils

# Interpolation fill options
method_options = ["zero", "null", "bfill", "ffill", "linear"]


class Interpolation:
    def __init__(self, is_resampled: bool):
        self.is_resampled = is_resampled

    def __validate_fill(self, method: str) -> None:
        """
        Validate if the fill provided is within the allowed list of values.

        :param fill: Fill type e.g. "zero", "null", "bfill", "ffill", "linear"
        """
        if method not in method_options:
            raise ValueError(
                f"Please select from one of the following fill options: {method_options}"
            )

    def __validate_col(
        self,
        df: DataFrame,
        partition_cols: Optional[List[str]],
        target_cols: List[str],
        ts_col: str,
        ts_col_dtype: Optional[str] = None,  # NB: added for testing purposes only
    ) -> None:
        """
        Validate if target column exists and is of numeric type, and validates if partition column exists.

        :param df: DataFrame to be validated
        :param partition_cols: Partition columns to be validated
        :param target_col: Target column to be validated
        :param ts_col: Timestamp column to be validated
        """

        if partition_cols is not None:
            for column in partition_cols:
                if column not in str(df.columns):
                    raise ValueError(
                        f"Partition Column: '{column}' does not exist in DataFrame."
                    )
        for column in target_cols:
            if column not in str(df.columns):
                raise ValueError(
                    f"Target Column: '{column}' does not exist in DataFrame."
                )

        if ts_col not in str(df.columns):
            raise ValueError(
                f"Timestamp Column: '{ts_col}' does not exist in DataFrame."
            )

        if ts_col_dtype is None:
            ts_col_dtype = df.select(ts_col).dtypes[0][1]
        if ts_col_dtype != "timestamp":
            raise ValueError("Timestamp Column needs to be of timestamp type.")

    def __calc_linear_spark(
        self, df: DataFrame, ts_col: str, target_col: str
    ) -> DataFrame:
        """
        Native Spark function for calculating linear interpolation on a DataFrame.

        :param df: prepared dataframe to be interpolated
        :param ts_col: timeseries column name
        :param target_col: column to be interpolated
        """
        interpolation_expr = f"""
        case when is_interpolated_{target_col} = false then {target_col}
            when {target_col} is null then
            (next_null_{target_col} - previous_{target_col})
            /(unix_timestamp(next_timestamp_{target_col})-unix_timestamp(previous_timestamp_{target_col}))
            *(unix_timestamp({ts_col}) - unix_timestamp(previous_timestamp_{target_col}))
            + previous_{target_col}
        else
            (next_{target_col}-{target_col})
            /(unix_timestamp(next_timestamp)-unix_timestamp(previous_timestamp))
            *(unix_timestamp({ts_col}) - unix_timestamp(previous_timestamp))
            + {target_col}
        end as {target_col}
        """

        # remove target column to avoid duplication during interpolation expression
        cols: List[str] = df.columns
        cols.remove(target_col)
        interpolated: DataFrame = df.selectExpr(*cols, interpolation_expr)
        # Preserve column order
        return interpolated.select(*df.columns)

    def _is_valid_method_for_column(
        self, series: DataFrame, method: str, col_name: str
    ) -> bool:
        """
        zero and linear interpolation are only valid for numeric columns
        """
        if method in ["linear", "zero"]:
            return isinstance(series.schema[col_name].dataType, NumericType)
        else:
            return True

    def __interpolate_column(
        self,
        series: DataFrame,
        ts_col: str,
        target_col: str,
        method: str,
    ) -> DataFrame:
        """
        Apply interpolation to column.

        :param series: input DataFrame
        :param ts_col: timestamp column name
        :param target_col: column to interpolate
        :param method: interpolation function to fill missing values
        """

        if not self._is_valid_method_for_column(series, method, target_col):
            raise ValueError(
                f"Interpolation method '{method}' is not supported for column "
                f"'{target_col}' of type '{series.schema[target_col].dataType}'. "
                f"Only NumericType columns are supported."
            )

        output_df: DataFrame = series

        # create new column for if target column is interpolated
        flag_expr = f"""
        CASE WHEN {target_col} is null and is_ts_interpolated = false THEN true
             WHEN is_ts_interpolated = true THEN true
        ELSE false
        END AS is_interpolated_{target_col}
        """
        output_df = output_df.withColumn(
            f"is_interpolated_{target_col}", sfn.expr(flag_expr)
        )

        # Handle zero fill
        if method == "zero":
            output_df = output_df.withColumn(
                target_col,
                sfn.when(
                    sfn.col(f"is_interpolated_{target_col}") == False,  # noqa: E712
                    sfn.col(target_col),
                ).otherwise(sfn.lit(0)),
            )

        # Handle null fill
        if method == "null":
            output_df = output_df.withColumn(
                target_col,
                sfn.when(
                    sfn.col(f"is_interpolated_{target_col}") == False,  # noqa: E712
                    sfn.col(target_col),
                ).otherwise(None),
            )

        # Handle forward fill
        if method == "ffill":
            output_df = output_df.withColumn(
                target_col,
                sfn.when(
                    sfn.col(f"is_interpolated_{target_col}") == True,  # noqa: E712
                    sfn.col(f"previous_{target_col}"),
                ).otherwise(sfn.col(target_col)),
            )
        # Handle backwards fill
        if method == "bfill":
            output_df = output_df.withColumn(
                target_col,
                # Handle case when subsequent value is null
                sfn.when(
                    (sfn.col(f"is_interpolated_{target_col}") == True)  # noqa: E712
                    & (
                        sfn.col(f"next_{target_col}").isNull()
                        & (sfn.col(f"{ts_col}_{target_col}").isNull())
                    ),
                    sfn.col(f"next_null_{target_col}"),
                ).otherwise(
                    # Handle standard backwards fill
                    sfn.when(
                        sfn.col(f"is_interpolated_{target_col}") == True,  # noqa: E712
                        sfn.col(f"next_{target_col}"),
                    ).otherwise(sfn.col(f"{target_col}"))
                ),
            )

        # Handle linear fill
        if method == "linear":
            output_df = self.__calc_linear_spark(
                output_df,
                ts_col,
                target_col,
            )

        return output_df

    def __generate_time_series_fill(
        self, df: DataFrame, partition_cols: Optional[List[str]], ts_col: str
    ) -> DataFrame:
        """
        Create additional timeseries columns for previous and next timestamps

        :param df: input DataFrame
        :param partition_cols: partition column names
        :param ts_col: timestamp column name
        """
        return df.withColumn(
            "previous_timestamp",
            sfn.col(ts_col),
        ).withColumn(
            "next_timestamp",
            sfn.lead(df[ts_col]).over(
                Window.partitionBy(*partition_cols).orderBy(ts_col)
            ),
        )

    def __generate_column_time_fill(
        self,
        df: DataFrame,
        partition_cols: Optional[List[str]],
        ts_col: str,
        target_col: str,
    ) -> DataFrame:
        """
        Create timeseries columns for previous and next timestamps for a specific target column

        :param df: input DataFrame
        :param partition_cols: partition column names
        :param ts_col: timestamp column name
        :param target_col: target column name
        """
        window = Window
        if partition_cols is not None:
            window = Window.partitionBy(*partition_cols)

        return df.withColumn(
            f"previous_timestamp_{target_col}",
            sfn.last(sfn.col(f"{ts_col}_{target_col}"), ignorenulls=True).over(
                window.orderBy(ts_col).rowsBetween(Window.unboundedPreceding, 0)
            ),
        ).withColumn(
            f"next_timestamp_{target_col}",
            sfn.last(sfn.col(f"{ts_col}_{target_col}"), ignorenulls=True).over(
                window.orderBy(sfn.col(ts_col).desc()).rowsBetween(
                    Window.unboundedPreceding, 0
                )
            ),
        )

    def __generate_target_fill(
        self,
        df: DataFrame,
        partition_cols: Optional[List[str]],
        ts_col: str,
        target_col: str,
    ) -> DataFrame:
        """
        Create columns for previous and next value for a specific target column

        :param df: input DataFrame
        :param partition_cols: partition column names
        :param ts_col: timestamp column name
        :param target_col: target column name
        """
        window = Window

        if partition_cols is not None:
            window = Window.partitionBy(*partition_cols)
        return (
            df.withColumn(
                f"previous_{target_col}",
                sfn.last(df[target_col], ignorenulls=True).over(
                    window.orderBy(ts_col).rowsBetween(Window.unboundedPreceding, 0)
                ),
            )
            # Handle if subsequent value is null
            .withColumn(
                f"next_null_{target_col}",
                sfn.last(df[target_col], ignorenulls=True).over(
                    window.orderBy(sfn.col(ts_col).desc()).rowsBetween(
                        Window.unboundedPreceding, 0
                    )
                ),
            ).withColumn(
                f"next_{target_col}",
                sfn.lead(df[target_col]).over(window.orderBy(ts_col)),
            )
        )

    def interpolate(
        self,
        tsdf: t_tsdf.TSDF,
        ts_col: str,
        partition_cols: Optional[List[str]],
        target_cols: List[str],
        freq: Optional[str],
        func: Optional[Union[Callable | str]],
        method: str,
        show_interpolated: bool,
        perform_checks: bool = True,
    ) -> DataFrame:
        """
        Apply interpolation function.

        :param tsdf: input TSDF
        :param ts_col: timestamp column name
        :param target_cols: numeric columns to interpolate
        :param partition_cols: partition columns names
        :param freq: frequency at which to sample
        :param func: aggregate function used for sampling to the specified interval
        :param method: interpolation function usded to fill missing values
        :param show_interpolated: show if row is interpolated?
        :param perform_checks: calculate time horizon and warnings if True (default is True)
        :return: DataFrame containing interpolated data.
        """
        # Validate input parameters
        self.__validate_fill(method)
        self.__validate_col(tsdf.df, partition_cols, target_cols, ts_col)

        if freq is None:
            raise ValueError("freq cannot be None")

        if func is None:
            raise ValueError("func cannot be None")

        if callable(func):
            raise ValueError("func must be a string")

        # Convert Frequency using resample dictionary
        parsed_freq = t_resample.checkAllowableFreq(freq)
        period, unit = parsed_freq[0], parsed_freq[1]
        freq = f"{period} {t_resample.freq_dict[unit]}"  # type: ignore[literal-required]

        # Throw warning for user to validate that the expected number of output rows is valid.
        if perform_checks:
            t_utils.calculate_time_horizon(tsdf.df, ts_col, freq, partition_cols)

        # Only select required columns for interpolation
        input_cols: List[str] = [ts_col, *target_cols]
        if partition_cols is not None:
            input_cols += [*partition_cols]

        sampled_input: DataFrame = tsdf.df.select(*input_cols)

        if self.is_resampled is False:
            # Resample and Normalize Input
            sampled_input = tsdf.resample(
                freq=freq, func=func, metricCols=target_cols
            ).df

        # Fill timeseries for nearest values
        time_series_filled = self.__generate_time_series_fill(
            sampled_input, partition_cols, ts_col
        )

        # Generate surrogate timestamps for each target column
        # This is required if multuple columns are being interpolated and may contain nulls
        add_column_time: DataFrame = time_series_filled
        for column in target_cols:
            add_column_time = add_column_time.withColumn(
                f"{ts_col}_{column}",
                sfn.when(sfn.col(column).isNull(), None).otherwise(sfn.col(ts_col)),
            )
            add_column_time = self.__generate_column_time_fill(
                add_column_time, partition_cols, ts_col, column
            )

        # Handle edge case if last value (latest) is null
        edge_filled = add_column_time.withColumn(
            "next_timestamp",
            sfn.when(
                sfn.col("next_timestamp").isNull(),
                sfn.expr(f"{ts_col}+ interval {freq}"),
            ).otherwise(sfn.col("next_timestamp")),
        )

        # Fill target column for nearest values
        target_column_filled = edge_filled
        for column in target_cols:
            target_column_filled = self.__generate_target_fill(
                target_column_filled, partition_cols, ts_col, column
            )

        # Generate missing timeseries values
        exploded_series = target_column_filled.withColumn(
            f"new_{ts_col}",
            sfn.expr(
                f"explode(sequence({ts_col}, next_timestamp - interval {freq}, interval {freq} )) as timestamp"
            ),
        )
        # Mark rows that are interpolated if flag is set to True
        flagged_series: DataFrame = exploded_series

        flagged_series = (
            exploded_series.withColumn(
                "is_ts_interpolated",
                sfn.when(sfn.col(f"new_{ts_col}") != sfn.col(ts_col), True).otherwise(
                    False
                ),
            )
            .withColumn(ts_col, sfn.col(f"new_{ts_col}"))
            .drop(sfn.col(f"new_{ts_col}"))
        )

        # # Perform interpolation on each target column
        interpolated_result: DataFrame = flagged_series
        for target_col in target_cols:
            # Interpolate target columns
            interpolated_result = self.__interpolate_column(
                interpolated_result, ts_col, target_col, method
            )

            interpolated_result = interpolated_result.drop(
                f"previous_timestamp_{target_col}",
                f"next_timestamp_{target_col}",
                f"previous_{target_col}",
                f"next_{target_col}",
                f"next_null_{target_col}",
                f"{ts_col}_{target_col}",
            )

        # Remove non-required columns
        output: DataFrame = interpolated_result.drop(
            "previous_timestamp", "next_timestamp"
        )

        # Hide is_interpolated columns based on flag
        if show_interpolated is False:
            interpolated_col_names = ["is_ts_interpolated"]
            for column in target_cols:
                interpolated_col_names.append(f"is_interpolated_{column}")
            output = output.drop(*interpolated_col_names)

        return output